#!/usr/bin/php
<?
/*
 *
 * db-cfd.sh
 * 
 * (c) 2012 Felipe Eduardo Sanchez Diaz Duran
 * 
 * 
 * Uso:
 * 
 * db-cfd.sh <archivo> <RFC> <usuario CFD>
 * 
 * Lee el CFD <archivo> y lo registra en una base de datos.
 * 
 * El parametro <RFC> indica el RFC del contribuyente.
 *
 * <usuario CFD> es el nombre del usuario de ocfdweb.
 * Por ejemplo: asic@comprobantefiscal.mx
 * 
 */

/************************/
/*     Definiciones     */
/************************/

define ("RE_RFC", "/^[A-Za-z&]{3,4}[\- ]{0,1}\d{6}(?:[\- ]{0,1}[A-Za-z\d]{3})?$/");

define ("RET_OK", 0);
define ("RET_ERROR", 1);
define ("RET_DUPLICADO", 2);
define ("RET_V3_NO_UUID", 3);


/*************************/
/*       Variables       */
/*************************/

$database = "organizacfd";
$usuario_bd = "postgres";
$password_bd = "xxxxxxx";

/*************************/
/*       Funciones       */
/*************************/

/*
 * conectadb()
 *
 * Crea una conexion a nuestra base de datos
 */

function conectadb() {
   global $database, $usuario_bd, $password_bd;

   if (! pg_connect("dbname=$database user=$usuario_bd password=$password_bd") ) {
      print ("Error al conectar con la BD:" . pg_last_error());
      exit (RET_ERROR);
   }

   pg_set_client_encoding('LATIN1');
}

/*                                                                                                                                                  
 * sql_params()                                                                                                                                     
 *                                                                                                                                                  
 * Recibe un query parametrizado lo ejecuta y regresa los resultados, verificacion                                                                  
 * de errores.                                                                                                                                      
 */                                                                                                                                                 
                                                                                                                                                    
                                                                                                                                                    
function sql_params($query, $parametros) {                                                                                                          
                                                                                                                                                    
   $resultado = pg_query_params ($query, $parametros);                                                                                              
   if ( ! $resultado ) {                                                                                                                            
      print "Ocurrio un error SQL:" . pg_last_error(). "\n";                                  
      print "QUERY: $query\n";                                                                               
      exit (RET_ERROR);
   }                                                                                                                                                
                                                                                                                                                    
   return $resultado;                                                                                                                               
}                                                                                                                                                   


/*
 * DameRenglon()
 *
 * Recibe un resultado de query y devuelve el siguiente renglon en
 * el resultado. Es realmente una envoltura para no estar llamando
 * a pg_fetch_array siempre con los mismos argumentos.
 */

function DameRenglon($resultado) {
   return pg_fetch_array($resultado, NULL, PGSQL_ASSOC);
}



/*
 * valorcfd()
 *
 * Recibe una cadena XPath y devuelve una cadena con el contenido
 * del elemento solicitado.
 *
 * Evalua si es un CFD o un CFDi.
 */

function valorcfd ($xpathstr) {

   /* Solo conocemmos tres namespaces:
    *  - http://www.sat.gob.mx/cfd/3 (cfdi)
    *  - http://www.sat.gob.mx/cfd/2 (cfd, pero se usa con cadena vacia)
    *  - http://www.sat.gob.mx/cfd/1 (No hemos encontrado ninguno hasta ahora)
    */


   global $cfd;

   $namespaces = $cfd->getNameSpaces(FALSE);

   /* Identifiquemos si es cfd, cfdi o alguna otra cosa*/
   if       (in_array("http://www.sat.gob.mx/cfd/3", $namespaces ))  {
     $ns = "http://www.sat.gob.mx/cfd/3";
   } elseif (in_array("http://www.sat.gob.mx/cfd/2", $namespaces)) {
     $ns = "http://www.sat.gob.mx/cfd/2";
   } elseif (in_array("http://www.sat.gob.mx/cfd/1", $namespaces))  {
     $ns = "http://www.sat.gob.mx/cfd/1";
   } else {
     $ns = "";
   }

   /* El xpathstr siempre tiene un prefijo de namespace "ocfd".   */
   /* De acuerdo al namespace que el documento ha declarado y que */
   /* identificamos aca arriba vamos a traducirlo siempre         */
   /* al prefijo "ocfd".                                          */

   $cfd->registerXPathNamespace('ocfd', $ns);

   /* No se supone que haya elementos repetidos. En todo caso */
   /* si los hubiera solo regresaremos el primero.            */

   foreach ($cfd->xpath($xpathstr) as $elem) return $elem;

}

/*
 * NormalizaRFC()
 *
 * Elimina espacios en blanco y guiones del RFC y lo convierte
 * a mayusculas.
 *
 */

function NormalizaRFC($rfc) {

   return strtoupper (str_replace(array('-','_',' ', "\t"), NULL, $rfc));

}




/*************************/
/*    Bucle principal    */
/*************************/


print ("=== Procesando " . $argv[1] . " ===\n");

/***** Validamos los argumentos *****/

if ($argc < 4) {
   print ("Uso: " . basename($argv[0]) . " <archivo> <RFC> <usuario CFD>\n");
   exit (RET_ERROR);
}

/* Existe el archivo? */

if ( file_exists ( $argv[1]) == FALSE ) {
   printf ("El archivo %s no existe\n", $argv[1]);
   exit (RET_ERROR);
}

/* El RFC es valido? */

if ( ! preg_match ( RE_RFC, $argv[2] ) ) {
   printf ("El RFC no es valido\n");
   exit (RET_ERROR);
}

$cfdfile    = $argv[1];
$mirfc      = $argv[2];
$usuariocfd = $argv[3];

/* Carguemos el CFD como objeto XML */
$cfd = simplexml_load_file ( $argv[1] );


/* Extraigamos los campos que nos interesan */


$rfcemisor   = NormalizaRFC(valorcfd("//ocfd:Emisor/@rfc"));
$rs_emisor   = valorcfd("/ocfd:Comprobante/ocfd:Emisor/@nombre");
$rfcreceptor = NormalizaRFC(valorcfd("//ocfd:Receptor/@rfc"));
$rs_receptor = valorcfd("/ocfd:Comprobante/ocfd:Receptor/@nombre");
$folio       = valorcfd("/ocfd:Comprobante/@folio");
$serie       = valorcfd("/ocfd:Comprobante/@serie");
$fecha       = valorcfd("/ocfd:Comprobante/@fecha");
$version     = valorcfd("/ocfd:Comprobante/@version")?:1; // CFDv1 no incluye atributo @version




/* UUID para los que tienen un TimbreFiscalDigital */
if ( in_array ("http://www.sat.gob.mx/TimbreFiscalDigital", $namespaces = $cfd->getDocNameSpaces(TRUE))) {
   $cfd->registerXPathNamespace('tfd', "http://www.sat.gob.mx/TimbreFiscalDigital");
   $cfd->registerXPathNamespace('cfdi', "http://www.sat.gob.mx/cfd/3");
   $uuid     = strtoupper(valorcfd("/cfdi:Comprobante/cfdi:Complemento/tfd:TimbreFiscalDigital/@UUID"));
} else {
   $uuid     = NULL;
}

/* Un CFD v3.x sin UUID no es valido */
if ( preg_match("/^3(\.?\d+|\d*)$/", $version) && empty ($uuid) ) {
   printf ("El archivo %s no es valido: dice ser un CFDi version %s pero no contiene un UUID.", $cfdfile, $version);
   exit (RET_V3_NO_UUID);
}



/* El tipo de comprobante lo codificamos como un solo caracter */
switch ( strtolower (valorcfd("/ocfd:Comprobante/@tipoDeComprobante")) ) {
   case "ingreso":
      $tipo_comprobante = 'i';
      break;
   case "egreso":
      $tipo_comprobante = 'e';
      break;
   case "traslado":
      $tipo_comprobante = 't';
      break;
   default:
      $tipo_comprobante = '-';
}

/* Pa' la BD! */

conectadb();

/* Cambiamos los caracteres especiales de $usuariocfd por "_" */
/* y lo usamos como nombre del schema.                        */

$schema_usuario = str_replace(array('-','@','.'), '_', $usuariocfd);

printf ("rfcemisor x%sx, serie: x%sx, folio x%sx, uuid: x%sx\n", $rfcemisor, $serie, $folio, $uuid);

/* Construimos el query y el array de acuerdo a los valores que tenemos para */
/* serie, folio y uuid. SQL requiere que una prueba por un valor nulo sea    */
/* columna is NULL, asi que hay que ver cuales son nulos y cuales no y       */
/* construir el query y el array.                                            */

$cuenta_args = 1;
$query_args = array($rfcemisor);
$condiciones = "WHERE rfcemisor = $1";

if ( empty ($serie) ) {
   $condiciones .= " AND ( serie IS NULL OR serie = '' ) ";
} else {
   $condiciones .= " AND serie = $" . ++$cuenta_args;
   $query_args[] = $serie;
}

if ( empty ($folio) ) {
   $condiciones .= " AND ( folio IS NULL OR folio = '' ) ";
} else {
   $cuentaargs++;
   $condiciones .= " AND folio = $" . ++$cuenta_args;
   $query_args[] = $folio;
}

if ( empty ($uuid) ) {
   $condiciones .= " AND ( uuid IS NULL OR uuid = '' ) ";
} else {
   $cuentaargs++;
   $condiciones .= " AND uuid = $" . ++$cuenta_args;
   $query_args[] = $uuid;
}

//print ("query: " . "SELECT rfcemisor, serie, folio, uuid FROM $schema_usuario.cfds " . $condiciones);
//var_dump($query_args);

/* Verifiquemos si no tenemos ya un CFD registrado con estos datos */
$row = DameRenglon (sql_params ("SELECT rfcemisor, serie, folio, uuid
                                 FROM $schema_usuario.cfds " . $condiciones,
                                 $query_args));


if ( $row == FALSE ) {

   sql_params ("INSERT INTO $schema_usuario.cfds
                VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, default)",
                array ($serie,
                       $folio,
                       $rfcemisor,
                       $rfcreceptor,
                       $rs_emisor,
                       $rs_receptor,
                       $fecha,
                       $cfd->asXML(),
                       basename ($cfdfile),
                       strtolower ($tipo_comprobante),
                       $uuid,
                       $version) );

   printf ("archivo: %s, RFC: %s\n", $argv[1], $argv[2]);

} else {
   printf ("Ya existe un CFD registrado con estos datos: Emisor: %s, Serie: %s, Folio: %s, Fecha: %s, UUID: %s\n", $rfcemisor, $serie, $folio, $fecha, $uuid);
   exit (RET_DUPLICADO);
}



exit (RET_OK);

?>
